<?php declare(strict_types=1);

namespace Repositories;

use Carbon\Carbon;
use Database\Connection;
use Exceptions\TaskNotFoundException;
use Interfaces\TaskDao;
use Models\Task;
use PDO;

class TaskRepository implements TaskDao
{

    private PDO $conn;


    public function __construct(Connection $connection)
    {
        $this->conn = $connection->getConnection();
    }

    public function create(Task $task): int
    {
        $data = $task->persist();
        $stmt = $this->conn->prepare(
            "INSERT INTO Tasks (task_owner, created_at, updated_at, name, description, start, end, status_id) VALUES (?, ?, ?, ?, ?, ?, ?, ?)"
        );

        $stmt->bindParam(1, $data['task_owner']);
        $stmt->bindParam(2, $data['created_at']);
        $stmt->bindParam(3, $data['updated_at']);
        $stmt->bindParam(4, $data['name']);
        $stmt->bindParam(5, $data['description']);
        $stmt->bindParam(6, $data['start']);
        $stmt->bindParam(7, $data['end']);
        $stmt->bindParam(8, $data['status_id']);

        if ($stmt->execute()) {
            return (int) $this->conn->lastInsertId();
        }

        return 0;
    }

    public function readAll(): array
    {
        $stmt = $this->conn->query('SELECT * FROM Tasks');
        return $stmt->fetchAll(PDO::FETCH_ASSOC);
    }

    /**
     * @throws TaskNotFoundException
     */
    public function readById($id): Task
    {
        $stmt = $this->conn->prepare('SELECT * FROM Tasks WHERE id = ?');
        $stmt->execute([$id]);

        if (!$data = $stmt->fetch(PDO::FETCH_ASSOC))
        {
            throw new TaskNotFoundException('Task not found.');
        }

        return Task::fromArray($data);
    }

    public function update(Task $task): bool
    {
        $stmt = $this->conn->prepare(
          "UPDATE Tasks SET task_owner = ?, updated_at = ?, name = ?, description = ?, start = ?, end = ?, status_id = 0 WHERE id = ?"
        );

        return $stmt->execute([
            1, // Mock user ID for now, we haven't implemented authentication yet
            Carbon::now(),
            $task->getName(),
            $task->getDescription(),
            $task->getStart(),
            $task->getEnd(),
            $task->getStatusId(),
            $task->getTaskId()
        ]);
    }

    // fields = ['name', 'description']

    // TODO: Abstract this implementation
    // TODO: Check for injection vectors; fields might be used to pollute the query. Fields aren't bindParam'able.
    public function updateSelective(Task $task, array $fields): bool
    {
        if (empty($fields))
        {
            return false;
        }

        $setClause = [];
        $params = [];


        foreach ($fields as $field)
        {
            // TODO: use the Dirty implementation. We can chop off the $fields param and still not trust that data.
            // Consider using a Trait to control this behavior (dirty or dynamic get: e.g Uses DirtyFields or Uses ManualUpdates)
            // we're discarding invalid properties here; if they give us --name;, this check will fail.
            // essentially, the Model is serving as our array of allowed fields.
            if (property_exists($task,  snakeToCamel($field)))
            {
                $setClause[] = camel_to_snake($field) . " = ?"; // SET name = ?, SET updatedAt = ? (fixed by camel_to_snake)

                if ($field == 'status_id')
                {
                    $params[] = $task->getStatusId()->value;
                    continue;
                }
                $params[] = $task->{'get' . ucfirst(snakeToCamel($field))}();

            }
        }

        if (empty($setClause))
        {
            return false;
        }

        // we might have to update this and set the id later directly from the model
        $params[] = $task->getTaskId();

        $query = "UPDATE Tasks SET " . implode(', ', $setClause) . " WHERE id = ?";
        $stmt = $this->conn->prepare($query);

        return $stmt->execute($params);
    }

    public function delete(Task $task): bool
    {
        $stmt = $this->conn->prepare('DELETE FROM Tasks WHERE id = ?');
        return $stmt->execute([$task->getTaskId()]);
    }
}